
require 'rexml/document'
require 'enumerator'
require 'amrita2/util'

module REXML
  class Child 
    def amrita_compile_main_code(cg)
      cg.put_constant(to_s)
    end
  end

  class Parent < Child
    def amrita_compile_main_code(cg)
      each_child do |c|
        c.amrita_compile_main_code(cg)
      end
    end
  end

  class Element < Parent
    attr_accessor :amrita_spec
    def amrita_compile_main_code(cg)
      cg.compile_element_main_code(self) do
        super
      end
    end

    def amrita_compile_sub_code(cg)
      cg.compile_element_sub_code(self) do
        elements.each do |e|
          e.amrita_compile_sub_code(cg)
        end
      end
    end
  end
end

module Amrita2
  module Core
    class CodeGenerator
      attr_accessor :current_stream

      def initialize(stream = "get_mainstream")
        @current_stream = stream
        init_results
        init_strbuf
        @current_indent = 0
        @dynamic_element_stack = []
        @varcnt=0
      end

      def put_constant(msg)
        @strbuf.concat msg
      end

      def flush
        if @strbuf.size > 0
          @results << [@current_indent, "#{current_stream} << " + @strbuf.inspect]
          init_strbuf
        end
      end

      def put_expression(exp)
        code("#{current_stream} << #{exp}")
      end

      def code(l)
        flush
        @results << [@current_indent, l]
      end

      def new_var
        @varcnt += 1
        "v#{@varcnt}"
      end

      def results
        flush
        @results.collect do |level, line|
          "  " * level + line
        end
      end

      def if_(cond)
        code("if #{cond}")
        level_up do
          yield
        end
        code("end")
      end

      def else_
        level_down do
          code("else")
        end
        yield
      end

      def level_up
        flush
        @current_indent += 1
        yield
      ensure
        flush
        @current_indent -= 1
      end

      def level_down
        flush
        @current_indent -= 1
        yield
      ensure
        flush
        @current_indent += 1
      end

      def define_method(name, *params)
        if params.size == 0
          code("def #{name}")
        else
          code("def #{name}(#{params.join(',')})")
        end
        level_up do
          yield
        end
        code("end")
      end

      def define_module(name)
        code("module #{name}")
        level_up do
          yield
        end
        code("end")
      end

      def convert(encto, encfrom, *args)
        a = args.join(",")
        code("#{a} = *Iconv::iconv(#{encto.inspect}, #{encfrom.inspect}, #{a})")
      end

      def convert_hash(encto, encfrom, hash)
        code("#{hash}.each do |k, v|")
        level_up do
          code("v = *Iconv::iconv(#{encto.inspect}, #{encfrom.inspect}, v)")
          code("#{hash}[k] = v")
        end
        code("end")
      end

      def compile_element_main_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_main_code(element, self, &block)
        else
          generate_static_element_code(element, &block)
        end
      end

      def compile_element_sub_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_sub_code(element, self, &block) 
        else
          yield
        end
      end

      def init_results
        @results = []
      end

      def init_strbuf
        @strbuf = ""
      end

      def generate_static_element_code(e, option={}, &block)
        if e.name and e.name.size > 0
          unless option[:delete_tag]
            a = []
            a << e.name 
            if (option[:delete_id])
              a += e.attributes.enum_for(:each_attribute).find_all do |attr|
                attr.name != option[:key_attr_name].to_s
              end.collect do |attr| 
                attr.to_string 
              end
            else
              a += e.attributes.enum_for(:each_attribute).collect { |attr| attr.to_string }
            end
            put_constant("<#{a.join(' ')}>")
          end
          yield
          put_constant("</#{e.name}>") unless option[:delete_tag]
        else
          yield
        end
      end

      def new_context(new_stream,  &block)
        code("new_context(#{new_stream}) do")
        level_up do
          yield
        end
        code("end")
      end

#       def new_context(new_stream,  &block)
#         save_stream = @current_stream
#         code("new_context(#{new_stream}) do")
#         var = new_var
#         code("#{var} = get_mainstream")
#         @current_stream = var
#         level_up do
#           yield
#         end
#         code("end")
#         @current_stream = save_stream
#       end

      def dynamic_element_push(de)
        @dynamic_element_stack.unshift de
      end

      def dynamic_element_top
        @dynamic_element_stack.first
      end

      def dynamic_element_pop
        @dynamic_element_stack.shift
      end
    end

    class Matcher
      def find(e)
        return e if match(e)
        e.elements.each do |c|
          x = find(c)
          return x if x
        end
        nil
      end
    end

    class AttrMatcher < Matcher
      attr_reader :sym
      def initialize(sym, val)
        @sym = sym
        @sym_s = sym.to_s
        @val = val
      end

      def match(e)
        if @val != '*'
          @val == e.attributes[@sym_s]
        else
          e.attributes[@sym_s]
        end
      end
    end

    class IDMatcher < AttrMatcher
      def initialize(val)
        super(:id, val.to_s)
      end
    end

    class TagMatcher < Matcher
      def initialize(tag)
        @tag = tag.to_s
      end

      def match(e)
        e.name == @tag
      end
    end

    class AnyMatcher < Matcher
      def match(e)
        true
      end
    end

    module Runtime
      def new_element(tag, attrs={})
        ret = REXML::Element.new(tag.to_s)
        attrs.each do |k,v|
          ret.attributes[k.to_s] = v.to_s
        end
        ret
      end
      
      def start_tag(e, out="")
        out << "<#{e.expanded_name}"
        e.attributes.each_attribute do |attr|
          next if attr.value == ""
          out << " "
          attr.write(out)  
        end unless e.attributes.empty?
        out << ">"
        out
      end

      def end_tag(e, out="")
        out << "</#{e.expanded_name}>"
        out
      end

      class Context 
        attr_reader :main_stream
        
        def initialize(stream)
          @main_stream = stream
          @sub_stream = {}
        end

        def sub_stream(key)
          @sub_stream[key] ||= ""
        end

        def clear_substream(key)
          @sub_stream[key] = ""
        end
      end
      
      def context_stack
        Thread::current[:amrita_context_stack] ||= []
      end

      def current_context
        context_stack.first
      end

      def new_context(stream="", &block)
        context_stack.unshift Context.new(stream)
        yield
        if stream.kind_of?(String)
          SanitizedString[stream]
        else
          stream
        end
      ensure
        context_stack.shift
      end

      def get_mainstream
        current_context.main_stream
      end

      def get_substream(key)
        current_context.sub_stream(key)
      end

      def output_substream(key)
        c = current_context
        s = c.sub_stream(key)
        c.main_stream << SanitizedString[s]
        c.clear_substream(key)
      end
    end
    

    class DynamicElementSpec
      attr_accessor :cloned, :parent
      attr_reader :sym, :matcher, :children, :option, :assigned_elements

      def initialize(sym, parent, option={}, &block)
        case sym
        when Symbol
          @sym = sym
        when nil
          @sym = nil
        else
          @sym = sym.to_s.intern
        end
        @parent = parent
        @cloned = false
        @option, @children = option, []
        @direct_define = @option[:direct_define]
        setup_matcher
        setup_generator
        @in_setup = true
        self.instance_eval(&block) if block_given?
        @in_setup = false
        setup_extention if self.respond_to? :setup_extention
      end

      def clone
        new_spec = DynamicElementSpec.new(sym, nil, option)
        @children.each_with_index do |child, i|
          c = child.clone
          c.parent = new_spec
          new_spec.children[i] = c
        end
        new_spec.cloned = true
        new_spec
      end

      def assign_element(e)
        raise "can't assign not cloned dynamic spec" unless @cloned
        @assigned_elements ||= []
        @assigned_elements << e
      end

      def matched_many?
        @option[:match_many] and @assigned_elements.size > 1 
      end

      def first_match?(element)
        matched_many? and @assigned_elements.index(element) == 0
      end

      def name_for(element)
        if matched_many?
          i = @assigned_elements.index(element)
          "#{@sym}_#{i}".intern
        else
          @sym
        end
      end

      def dynamic_element(name, opt={}, &block)
        opt[:key_attr_name] = @option[:key_attr_name]
        opt[:direct_define] = @direct_define
        opt[:delete_id] = @option[:delete_id] if @option.has_key?(:delete_id)
        opt[:encoding] = @option[:encoding] if @option.has_key?(:encoding)
        @children << DynamicElementSpec.new(name, self, opt, &block)
      end

      def parts_element(name, opt={}, &block)
        opt[:parts_element] = true
        dynamic_element(name, opt, &block)
      end

      def dummy_element(name, opt={}, &block)
        opt[:dummy_element] = true
        dynamic_element(name, opt, &block)
      end

      
      def compile_element_main_code(element, cg, &block)
        return if @option[:dummy_element]
        @generator.generate_main(element, self, cg, &block)
      end

      def compile_element_sub_code(element, cg, &block)
        return if @option[:dummy_element]
        @generator.generate_sub(element, self, cg, &block)
      end

      def method_missing(sym, *args, &block)
        if @direct_define and @in_setup
          params = args[-1] || {}
          dynamic_element(sym, params, &block)
        else
          super
        end
      end

      def module_name
        sym.to_s.capitalize  
      end

      def output_stream
        @option[:output_stream] or sym
      end

      def register_parts(modname=nil, method_name=nil)
        if method_name
          if modname 
            modname = module_name + "::" + modname
          else
            modname = module_name
          end
        else
          method_name = sym.to_s 
        end
        @parent.register_parts(modname, method_name)
      end

      def get_methods
        h = {}
        children.find_all do |c|
          ret = h[c.sym]
          h[c.sym] = true
          ret != true
        end.collect {|c| c.sym }
      end

      private
      def setup_matcher
        if sym
          @matcher = @option[:matcher] || AttrMatcher.new(@option[:key_attr_name], sym.to_s)
        else
          @matcher = AnyMatcher.new
        end
      end

      def setup_generator
        if @option[:generator] 
          @generator = @option[:generator]
        else
          @generator = DefaultCodeGenerator.new
          @generator.extend ReplaceAttr if @option[:replace_attr]
        end
      end
    end

    class RootSpec < DynamicElementSpec
      attr_reader :parts
      def initialize(option={}, &block)
        option[:key_attr_name] = :id unless option[:key_attr_name]
        super(nil, nil, option, &block)
        @parts = []
      end

      def clone
        new_spec = RootSpec.new(option)
        @children.each_with_index do |child, i|
          c = child.clone
          c.parent = new_spec
          new_spec.children[i] = c
        end
        new_spec.cloned = true
        new_spec
      end

      def module_name
        "MainContext"
      end

      def create_template(e)
        case e
        when REXML::Document
          doc = e
          root = e.root
        else
          doc = REXML::Document.new(e)
          root = doc.root
        end

        ret = Object.new
        m = create_template_module(root, doc)
        ret.extend(m)
        ret
      end

      def create_template_module(e, doc=nil)
        m = Module.new
        result = create_template_module_text(e)
        puts result  if $DEBUG
        m.module_eval result
        if doc
          m.const_set(:XmlDecl, doc.xml_decl.to_s)
          m.const_set(:DocType, doc.doctype.to_s)
        end
        m
      end

      def create_template_module_text(e)
        root = self.clone
        specs = [root]
        traverse_and_search_match(specs, e)
        # not match check
        raise "SPEC #{specs.first.sym} was not matched" if specs.size > 0 

        cg = CodeGenerator.new
        cg.code("include Amrita2::Runtime")
        cg.code("extend Amrita2::Runtime")
        
        cg.define_method(:expand_template, :out, "&block") do
          define_expand_template(cg, e)
        end

        cg.define_module(module_name) do
          cg.code("Methods = #{get_methods.inspect}")
          cg.code("extend Amrita2::Runtime")
          define_sub_modules(cg, e)
        end

        root.parts.each do |modname, method_name|
          cg.define_method(method_name, "*args", "&block") do
            cg.code('new_context() do')
            cg.level_up do
              cg.code("MainContext::#{modname}::#{method_name}(*args, &block)")
              cg.code("output_substream(:#{method_name})")
            end
            cg.code("end")
          end
          cg.code("module_function :#{method_name}")
        end

        cg.define_method(:xml_decl) do
          cg.code("XmlDecl.dup")
        end

        cg.define_method(:doctype) do
          cg.code("DocType.dup")
        end

        cg.results.join("\n")
      end

      def define_expand_template(cg, e)
        cg.code("new_context(out) do")
        cg.level_up do
          cg.code("block.call(#{module_name or 'nil'}) if block_given? ")
          e.amrita_compile_main_code(cg)
        end
        cg.code("end")
      end

      def define_sub_modules(cg, e)
        e.amrita_compile_sub_code(cg)
      end

      def traverse_and_search_match(child_specs, element)
        s = child_specs.first
        return unless s
        if s.matcher.match(element)
          element.amrita_spec = s
          s.assign_element(element)
          child_specs.shift unless s.option[:match_many]
          specs = s.children.clone
          traverse_and_search_match(specs, element)

          # not match check
          raise "SPEC #{specs.first.sym} was not matched" if specs.size > 0 and ! specs.first.option[:match_many]

          return true
        else
          element.elements.each do |ce|
            ret = traverse_and_search_match(child_specs, ce)
          end
          return false
        end
      end

      def register_parts(modname, method_name)
        @parts << [modname, method_name]
      end
    end

    class DefaultCodeGenerator
      def generate_main(element, dyn_spec, cg, &block)
        if dyn_spec.option[:parts_element]
          dyn_spec.register_parts
        else
          sym = element.amrita_spec.sym
          if sym
            generate_main_code(element, dyn_spec, cg, sym.to_s, &block)
          else
            cg.generate_static_element_code(element, &block)
          end
        end
      end

      def generate_sub(element, dyn_spec, cg, &block)
        generate_dynamic_element_code(element, dyn_spec, cg, &block)
      end

      private
      def generate_main_code(element, dyn_spec, cg, sym)
        cg.code("output_substream(:#{sym})")
      end

      def generate_dynamic_element_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        if dyn_spec.children.size == 0 
          code_for_leaf_element(element, dyn_spec, cg, &block)
        else
          if sym
            code_for_not_leaf_element(element, dyn_spec, cg, &block)
          else
            yield
          end
        end
      end

      def code_for_leaf_element(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "val", "attrs={}") do
          enc = dyn_spec.option[:encoding]
          if enc
            cg.convert("utf-8", enc, "sym", "val") 
            cg.convert_hash("utf-8", enc, "attrs")
          end
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            code_for_element_and_scalar(element, dyn_spec, cg)
          end
        end
        cg.code("module_function :#{sym}")
      end

      def code_for_not_leaf_element(element, dyn_spec, cg, &block)
        if dyn_spec.matched_many? and dyn_spec.first_match?(element)
          make_selection_code(element, dyn_spec, cg, &block)
        end
        sym = dyn_spec.name_for(element)
        cg.define_method(sym, "attrs={}", "&block") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            code_for_element_with_child(element, dyn_spec, cg, &block)
          end
        end
        cg.code("module_function :#{sym}")
        if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
          cg.define_module(dyn_spec.module_name) do
            cg.code("Methods = #{dyn_spec.get_methods.inspect}")
            cg.code("extend Amrita2::Runtime")
            yield
          end
        end
      end

      def make_selection_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "no=0", "attrs={}", "&block") do
          cg.code("no = no % #{dyn_spec.assigned_elements.size}")
          cg.code("mname = (#{sym.to_s.inspect} + '_' + no.to_s).intern")
          cg.code("self.send(mname, attrs, &block)")
        end        
        cg.code("module_function :#{sym}")
      end

      def code_for_element_and_scalar(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.put_expression("val.amrita_sanitize")
        end
      end

      def code_for_element_with_child(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |c|
            c.amrita_compile_main_code(cg)
          end
          #element.elements.each do |e|
          #  e.amrita_compile_main_code(cg)
          #end
        end
      end

      def gen_element_with_attr(element,dyn_spec, cg, varname='e')
        a = element.attributes.collect do |k, v|
          next if dyn_spec.option[:delete_id] and k == dyn_spec.option[:key_attr_name].to_s
          "#{k.intern.inspect} => #{v.inspect}, "
        end
        default_value = "{ #{a.join(' ')}}"
        cg.code("a = #{default_value}")
        cg.code("attrs.each do |k,v|")
        cg.level_up do
          cg.code("a[k] = v")
        end
        cg.code("end")
        cg.code("#{varname} = new_element(#{element.name.inspect}, a)")
      end
    end
  end


  module ReplaceAttr
    def code_for_element_and_scalar(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.put_expression("val.amrita_sanitize")
        end
        cg.else_ do
          gen_element_with_attr(element, dyn_spec, cg)
          cg.code("e.delete_attribute(#{dyn_spec.option[:key_attr_name].to_s.inspect})") if dyn_spec.option[:delete_id]
          cg.put_expression("start_tag(e)")
          cg.put_expression("val.amrita_sanitize")
          cg.put_expression("end_tag(e)")
        end
      end
    end

    def code_for_element_with_child(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
        end
        cg.else_ do
          gen_element_with_attr(element, dyn_spec, cg)
          cg.put_expression("start_tag(e)")
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
          cg.put_expression("end_tag(e)")
        end
      end
    end

  end


  def self.make_default_spec(root, option={:delete_id=>true, :check_true=>true, :replace_attr=>true, :key_attr_name=>:id})
    s = Core::RootSpec.new(option.clone)
    make_default_spec1(root, option, [s])
    s
  end

  def self.make_default_spec1(e, option, spec_stack)
    key_attr = option[:key_attr_name].to_s
    id_ = e.attributes[key_attr]
    if id_
      sym = id_.intern
      s = Core::DynamicElementSpec.new(sym, nil, option.clone)
      s.parent = spec_stack.first
      spec_stack.first.children << s
      spec_stack.unshift(s)
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack)
      end
      spec_stack.shift
    else
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack)
      end
    end
  end


  module HtmlTag
    include Amrita2::Core
    HEAD = TagMatcher.new(:head)
    TITLE = TagMatcher.new(:title)
    BODY = TagMatcher.new(:BODY)
    TABLE = TagMatcher.new(:table)
    TR = TagMatcher.new(:tr)
    TH = TagMatcher.new(:th)
    TD = TagMatcher.new(:td)
    UL = TagMatcher.new(:ul)
    OL = TagMatcher.new(:ol)
    LI = TagMatcher.new(:li)
    DL = TagMatcher.new(:dl)
    DT = TagMatcher.new(:dt)
    DD = TagMatcher.new(:dd)
    A = TagMatcher.new(:a)
    PRE = TagMatcher.new(:pre)
    DIV = TagMatcher.new(:div)
    SPAN = TagMatcher.new(:span)
    P = TagMatcher.new(:p)
  end

  Runtime = Core::Runtime
  def self.define_template_spec(option={}, &block)
    Core::RootSpec.new(option, &block)
  end
end

