require 'stringio'
require 'rexml/document'
require 'pp'
require 'amrita2/util'

module Amrita2
  # from Ruby Recipe Book
  module TextSupport 
    def unindent(str, n)
      indent_re = /^ {0,#{n}}/
      str.map { |line|
        line.sub(indent_re, "")
      }.join("")
    end
    
    def unindent_auto(str)
      #unindent(str, minimum_indent(str))
      unindent(str, firstline_indent(str))
    end

    def firstline_indent(str)
      str.to_a[0].slice(/\A */).length
    end

    def minimum_indent(str)
      indents = str.map {|line|
        line.slice(/\A */).length
      }
      ((indents - [0]).min || 0)
    end
  end

  class SpecGenerator
    include TextSupport
    NL = "\n"

    attr_reader :sample_data

    def initialize(doc, opt={})
      @doc = doc
      @option = opt
      @option[:key_attr_name] ||= :id
      @idcnt = 0
      @indent = 0
      @sample_data = {} 
      @keys = []
      if @option[:keys]
        @option[:keys].each {|k| add_key(k)}
      end
    end

    def add_key(key)
      @keys << key
    end

    def convert(out)
      @keys = [ @option[:key_attr_name] ] if @keys.size == 0
      convert1(@doc.root, out)
    end

    def generate_spec
      key_attr = @option[:key_attr_name]
      key_attr_code = ""
      key_attr_code << ":key_attr_name=> #{key_attr.inspect}"
      convert(out="")
<<-END
Amrita2::define_template_spec(#{key_attr_code}) do
#{out}
end
END
    end
    
    def generate_opt
      ret = []
      ret << "t.amrita_id = #{@option[:key_attr_name].inspect}" if @option[:key_attr_name]
      ret.join("\n")
    end


    def generate_all(f=nil)
      spec = generate_spec
      sd = StringIO.new
      PP.pp(sample_data, sd)
      opt = generate_opt
      if f
<<-END
require 'amrita2/template'
include Amrita2
t = TemplateFile.new(#{f.inspect}) do
#{spec}
end
data = 
#{sd.string}
t.expand(STDOUT, data)
END
      else
<<-END
require 'amrita2/template'
include Amrita2
TEXT = #{@doc.to_s.inspect}
t = TemplateText.new(TEXT) do
#{spec}
end
data = 
#{sd.string}
#{opt}
t.expand(STDOUT, data)
END
      end
    end

    private

    
    def convert1(element, out)
      if is_leaf(element)
        convert_leaf(element, out)
      else
        convert_not_leaf(element, out)
      end
    end

    def get_key_value(e)
      a = e.attributes
      @keys.each do |k|
        v = a[k.to_s]
        return [k, v] if v
      end
      [nil, nil]
    end

    def is_leaf(e)
      e.elements.each do |ee|
        return false if has_key(ee)
      end
      true
    end

    def has_key(e)
      return true if get_key_value(e)[0]
      e.elements.each do |ee|
        return true if has_key(ee)
      end
      false
    end

    def generate_matcher(k, v)
      if @option[:key_attr_name] == k
        nil
      else
        "Amrita2::Core::AttrMatcher.new(#{k.inspect}, #{v.to_s.inspect})"
      end
    end

    def generate_dynspec(k, v, dummy)
      ret = ""
      unless dummy
        ret << "dynamic_element("
      else
        ret << "dummy_element("
      end

      ret << v.to_s.intern.inspect
      m = generate_matcher(k, v)
      if m
        ret << ", :matcher => " << m
      end
      ret << ")"
      ret
    end

    def convert_leaf(element, out)
      key, key_value = *get_key_value(element)
      if key_value
        key_value = key_value.intern 
        #comment = (element.text || element.collect { |e| e.to_s}.join)
        comment = Amrita2::SanitizedString[element.collect { |e| e.to_s}.join]
        first_time = false
        case @sample_data[key_value]
        when nil
          first_time = true
          @sample_data[key_value] = comment
        when Array
          @sample_data[key_value] << comment
        else
          @sample_data[key_value] = [@sample_data[key_value], comment]
        end
        s = generate_dynspec(key, key_value, (not first_time))
        if comment
          comment = comment.gsub("\n", " ").strip 
          s += " # #{comment[0..20]} "
        end
        out << indent_space << s << NL
      end
    end

    def convert_not_leaf(element, out)
      key, key_value = *get_key_value(element)
      if key_value
        key_value = key_value.intern 
        first_time = false
        case @sample_data[key_value]
        when nil
          first_time = true
          @sample_data[key_value] = {}
        when Array
          @sample_data[key_value] << {}
        else
          @sample_data[key_value] = [@sample_data[key_value], {}]
        end

        s = generate_dynspec(key, key_value, (not first_time)) + " do"
        out << indent_space << s << NL 

        save_data = @sample_data
        @sample_data = @sample_data[key_value]
        @sample_data = @sample_data[-1] if @sample_data.kind_of?(Array)
        @indent += 1
        element.elements.each do |e|
          convert1(e, out)
        end
        @indent -= 1
        @sample_data = save_data
        out << indent_space << "end" << NL 
      else
        element.elements.each do |e|
          convert1(e, out)
        end
      end
    end

    def indent_space
      "  " * @indent
    end

    def new_id
      @idcnt += 1
      "e#{@idcnt}".intern
    end
  end
end

if __FILE__ == $0
  include Amrita2
  ARGV.each do |fname|
    doc = REXML::Document.new(File.new(fname))
    g = SpecGenerator.new(doc)
    puts g.generate_all
  end
end
